// $Id: JobControl.cpp 628 2011-08-31 15:42:16Z ge $
/// \file JobControl.h
/// \brief contains the JobControl class
///
/// $Revision: 1.10 $
/// \author Gerald Weber <gweberbh@gmail.com
#define GBC_JOBCONTROL_CPP "$Id: JobControl.cpp 628 2011-08-31 15:42:16Z ge $"

#include "JobControl.h"
#include <iterator>
#include <iostream>
#include <fstream>
#include <sstream>
#include <asm/param.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>

using namespace gbc;

using namespace std;

bool JobControl::JobControlEnabled=true;

pid_t JobControl::Pid=getpid();

pid_t JobControl::ParentPid=(pid_t)0;

string JobControl::ExecutableFile=string();

string JobControl::Username=string();

char JobControl::State=' ';

long int JobControl::Pagesize=getpagesize();

JobControl::int_map_type    JobControl::status_int_map=JobControl::int_map_type();
JobControl::string_map_type JobControl::status_string_map=JobControl::string_map_type();

void JobControl::disable(void)
  {
  JobControlEnabled=false;
  }

void JobControl::enable(void)
  {
  JobControlEnabled=true;
  }

string JobControl::proc_path(void)
  {
  stringstream path;
  path << "/proc/" << Pid;
  string pt;
  pt=path.str();
  return pt;
  }

string JobControl::Proc_path=JobControl::proc_path();

string JobControl::statm_path(void)
  {
  string sp=proc_path()+"/statm";
  return sp;
  }

ifstream JobControl::StatmFile;

string JobControl::stat_path(void)
  {
  string sp=proc_path()+"/stat";
  return sp;
  }


std::ifstream JobControl::StatFile;

string JobControl::make_log_dir(void)
  {
  char* NOJOBCONTROL=getenv("NOJOBCONTROL");
  if ((NOJOBCONTROL != NULL) || (!JobControlEnabled)) return string();

  read_stat();
  string logdir;
  string logfile;
  char* HOME=getenv("HOME");//man:getenv
  char* JOBCONTROL_DIR=getenv("JOBCONTROL_DIR");
  char* JOBCONTROL_FILE=getenv("JOBCONTROL_FILE");
  if (HOME != NULL) logdir=HOME;
  if (JOBCONTROL_DIR != NULL) logdir=JOBCONTROL_DIR;
  logdir+="/.bioinf";
  mkdir(logdir.c_str(),  S_IXUSR | S_IRUSR | S_IWUSR | S_IRGRP);
  if (JOBCONTROL_FILE != NULL) logfile=JOBCONTROL_FILE;
  else                         logfile=ExecutableFile;
  logdir+="/" + logfile;
  mkdir(logdir.c_str(),  S_IXUSR | S_IRUSR | S_IWUSR | S_IRGRP);
  return logdir;
  }

string JobControl::LogDir=string();

std::string JobControl::read_proc_file(std::string filename)
  {
  char st[256];
  ifstream proc_file;
  proc_file.open(filename.c_str());
  proc_file.getline(st,256);
  return string(st);
  }

string JobControl::Hostname=JobControl::read_proc_file("/proc/sys/kernel/hostname");
string JobControl::Osrelease=JobControl::read_proc_file("/proc/sys/kernel/osrelease");

string JobControl::log_path(string ext)
  {
  char* NOJOBCONTROL=getenv("NOJOBCONTROL");
  if ((NOJOBCONTROL != NULL) || (!JobControlEnabled)) return string();

  if (LogDir.empty()) LogDir=make_log_dir();
  read_stat();
  stringstream path;
  char* JOBCONTROL_FILE=getenv("JOBCONTROL_FILE");
  if (JOBCONTROL_FILE != NULL) 
    path << LogDir << "/" << JOBCONTROL_FILE  << ext;
  else 
    path << LogDir << "/" << Hostname << "."  << Pid << ext;
  string pts=path.str();
  return pts;
  }

std::string JobControl::LogFileName=string();

std::ofstream JobControl::LogFile;

std::string JobControl::StatusFileName=string();

std::ofstream JobControl::StatusFile;

long int JobControl::Pages=0L;
long int JobControl::MaxPages=0L;

long int JobControl::Cputime=0;

JobControl::JobControl(void)
  {
  }

long int JobControl::read_meminfo(void)
  {
  ifstream meminfo;
  meminfo.open("/proc/meminfo");
  char instring[254];
  long int mpm=0;
  if (Osrelease.compare(0,4,"2.4.",4)==0)
    {
    meminfo.getline(instring,254); //skip this    -> total:    used:    free:  shared: buffers:  cached:
    meminfo.getline(instring,254,':'); //skip until-> Mem:
    meminfo >> mpm;
    mpm/=Pagesize;
    meminfo.close();
    }
  if (Osrelease.compare(0,4,"2.6.",4)==0)
    {
    meminfo.getline(instring,254,':'); //skip until-> Mem:
    meminfo >> mpm;
    long int pkb=Pagesize/1024;
    mpm/=pkb; //this is the same as multiplying by 1024 and the dividing by the Pagesize
    meminfo.close();
    }
  return mpm;
  }

long int JobControl::MaxPhysicalMemory=read_meminfo();

int JobControl::physical_memory_in_use(void)
  {
  return (pages()*1000)/MaxPhysicalMemory;
  }

void JobControl::read_stat(void)
  {
  if (!StatFile.is_open()) StatFile.open(JobControl::stat_path().c_str());
  //comments from man:proc(5)
  // pid %d The process id.
  long int pid; StatFile >> pid; //ignored as pid is obtained from startup

  //comm %s The filename of the executable, in parentheses. This is visible whether or not the executable is swapped out.
  StatFile >> ExecutableFile;
  ExecutableFile.erase(ExecutableFile.find("("),1);
  ExecutableFile.erase(ExecutableFile.find(")"),1);
  
  //state %c One character from the string "RSDZTW" where R is running, S is sleeping in an interruptible wait, D is waiting in uninterruptible disk sleep, Z is zombie, T is traced or stopped (on a signal), and W is paging.
  StatFile >> State;

  //ppid %d The PID of the parent.
  StatFile >> ParentPid;

  //pgrp %d The process group ID of the process.
  long int pgrp; StatFile >> pgrp;

  //session %d The session ID of the process.
  long int session; StatFile >> session;

  //tty_nr %d The tty the process uses.
  long int tty_nr; StatFile >> tty_nr;

  //tpgid %d The process group ID of the process which currently owns the tty that the process is connected to.
  long int tpgid; StatFile >> tpgid;

  //flags %lu The flags of the process. The math bit is decimal 4, and the traced bit is decimal 10.
  unsigned long int flags; StatFile >> flags;

  //minflt %lu The number of minor faults the process has made which have not required loading a memory page from disk.
  unsigned long int minflt; StatFile >> minflt;

  //cminflt %lu The number of minor faults that the process and its children have made.
  unsigned long int cminflt; StatFile >> cminflt;

  //majflt %lu The number of major faults the process has made which have required loading a memory page from disk.
  unsigned long int majflt; StatFile >> majflt;

  //cmajflt %lu The number of major faults that the process and its children have made.
  unsigned long int cmajflt; StatFile >> cmajflt;

  //utime %lu The number of jiffies that this process has been scheduled in user mode.
  StatFile >> Cputime;

  //stime %lu The number of jiffies that this process has been scheduled in kernel mode.
  unsigned long int stime; StatFile >> stime; Cputime+=stime;

  //cutime %ld The number of jiffies that this process and its children have been scheduled in user mode.
  unsigned long int cutime; StatFile >> cutime;

  //cstime %ld The number of jiffies that this process and its children have been scheduled in kernel mode.
  unsigned long int cstime; StatFile >> cstime;

  //priority %ld The standard nice value, plus fifteen. The value is never negative in the kernel.
  //nice %ld The nice value ranges from 19 (nicest) to -19 (not nice to others).
  //0 %ld This value is hard coded to 0 as a placeholder for a removed field.
  //itrealvalue %ld The time in jiffies before the next SIGALRM is sent to the process due to an interval timer.
  //starttime %lu The time in jiffies the process started after system boot.
  //vsize %lu Virtual memory size in bytes.
  //rss %ld Resident Set Size: number of pages the process has in real memory, minus 3 for administrative purposes. This is just the pages which count towards text, data, or stack space. This does not include pages which have not been demand-loaded in, or which are swapped out.
  //rlim %lu Current limit in bytes on the rss of the process (usually 4294967295 on i386).
  //startcode %lu The address above which program text can run.
  //endcode %lu The address below which program text can run.
  //startstack %lu The address of the start of the stack.
  //kstkesp %lu The current value of esp (stack pointer), as found in the kernel stack page for the process.
  //kstkeip %lu The current EIP (instruction pointer).
  //signal %lu The bitmap of pending signals (usually 0).
  //blocked %lu The bitmap of blocked signals (usually 0, 2 for shells).
  //sigignore %lu The bitmap of ignored signals.
  //sigcatch %lu The bitmap of catched signals.
  //wchan %lu  This is the "channel" in which the process is waiting. It is the address of a system call, and can be looked up in a namelist if you need a textual name. (If you have an up-to-date /etc/psdatabase, then try ps -l to see the WCHAN field in action.)
  //nswap %lu Number of pages swapped - not maintained.
  //cnswap %lu Cumulative nswap for child processes.
  //exit_signal %d Signal to be sent to parent when we die.
  //processor %d CPU number last executed on.

  StatFile.seekg((streampos)0,ios::beg);
  }


long int JobControl::pages(void)
  {
  long int pg;
  if (!StatmFile.is_open()) StatmFile.open(statm_path().c_str());
  StatmFile >> pg >> pg; StatmFile.seekg((streampos)0,ios::beg);
  Pages=pg;
  if (Pages > MaxPages) MaxPages=Pages;
  return Pages;
  }

long int JobControl::diff_pages(void)
  {
  long int prev=Pages;
  return pages()-prev;
  }

long int JobControl::max_pages(void)
  {
  pages();
  return MaxPages;
  }

long int JobControl::memory(void)
  {
  return pages()*Pagesize;
  }

long int JobControl::diff_memory(void)
  {
  return diff_pages()*Pagesize;
  }

long int JobControl::max_memory(void)
  {
  return max_pages()*Pagesize;
  }

long int JobControl::cputime(void)
  {
  read_stat();
  return Cputime;
  }

long int JobControl::diff_cputime(void)
  {
  long int prev=Cputime;
  return cputime()-prev;
  }

void JobControl::write_log(string event)
  {
  char* NOJOBCONTROL=getenv("NOJOBCONTROL");
  if ((NOJOBCONTROL != NULL) || (!JobControlEnabled)) return;

  if (LogFileName.empty()) LogFileName=log_path(".log");
  if (!LogFile.is_open()) LogFile.open(LogFileName.c_str(),ios_base::app);
  if (event != string()) LogFile << event;
  LogFile << " " << cputime() << " " << diff_memory() << endl;
  LogFile.close();
  }

static time_t start_time=time(NULL);

void JobControl::write_status(void)
  {
  char* NOJOBCONTROL=getenv("NOJOBCONTROL");
  if ((NOJOBCONTROL != NULL) || (!JobControlEnabled)) return;

  read_stat();
  time_t nw=time(NULL);
  if (StatusFileName.empty()) StatusFileName=log_path(".status");
  if (!StatusFile.is_open()) StatusFile.open(StatusFileName.c_str(),ios_base::trunc);
  StatusFile << "[Process Information]" << endl;
  StatusFile << "Pid:" << Pid << endl;
  StatusFile << "Hostname:" << Hostname << endl;
  StatusFile << "Operating system release:" << Osrelease << endl;
  StatusFile << "Memory (b):" << MaxPages*Pagesize << endl;
  StatusFile << "Physical Memory (b):" << MaxPhysicalMemory*Pagesize << endl;
  StatusFile << "CPU time (cs):" << Cputime << endl;
  StatusFile << "Elapsed time (s): " << difftime(nw,start_time) << endl;
  StatusFile << "LogFile:" << JobControl::LogFileName << endl;
  StatusFile << "Started (local time):" << ctime(&start_time) << endl;
  StatusFile << "Updated (local time):" << ctime(&nw) << endl;
  int_map_type::iterator bt;
  StatusFile << "[Bool and Integer Flags]" << endl;
  for(bt=status_int_map.begin(); bt != status_int_map.end(); bt++)
    StatusFile << bt->first << ":" << bt->second << endl;
  string_map_type::iterator st;
  StatusFile << "[String Flags]" << endl;
  for(st=status_string_map.begin(); st != status_string_map.end(); st++)
    StatusFile << st->first << ":" << st->second << endl;
  StatusFile.close();
  }

void JobControl::status(string desc, int stat)
  {
  status_int_map[desc]=stat;
  write_status();
  }

void JobControl::status(string desc, string stat)
  {
  status_string_map[desc]=stat;
  write_status();
  }

JobControl::version_set_type JobControl::cvs_info =JobControl::version_set_type();

bool JobControl::cvs(std::string header,std::string module)
  {
  cvs_info.insert(header);
  if (module!=std::string()) cvs_info.insert(module);
  return true;
  }

bool CVS_info=JobControl::cvs(GBC_JOBCONTROL_H,GBC_JOBCONTROL_CPP);

ostream& JobControl::cvs(ostream& fl)
  {
  std::copy(cvs_info.begin(),cvs_info.end(),std::ostream_iterator<std::string>(fl,"\n"));
  return fl;
  }


void JobControl::summary(void)
  {
  char* NOJOBCONTROL=getenv("NOJOBCONTROL");
  if ((NOJOBCONTROL != NULL) || (!JobControlEnabled)) return;

  if (LogFileName.empty()) LogFileName=log_path(".log");
  if (!LogFile.is_open()) LogFile.open(LogFileName.c_str(),ios_base::app);
  summary(LogFile); 
  LogFile << endl;
  cvs(LogFile);
  LogFile.close();
  }

ostream& JobControl::summary(ostream& fl)
  {
  char* NOJOBCONTROL=getenv("NOJOBCONTROL");
  if ((NOJOBCONTROL != NULL) || (!JobControlEnabled)) return fl;

  time_t nw=time(NULL);
  fl << "Summary-> Cputime: "   << cputime()/100.0 << " s "
     << " Elapsed time: "       << difftime(nw,start_time) << " s "
     << " seconds max memory: " << max_memory()/1024 << " kb";
  return fl;
  }

ostream& JobControl::show_file_names(ostream& fl)
  {
  time(NULL);
  fl << "Log file:    " << LogFileName.c_str() << std::endl;
  fl << "Status file: " << StatusFileName.c_str() << std::endl;
  return fl;
  }

