// $Id: ReferenceSet.h 697 2012-11-05 15:57:05Z ge $
/// \file ReferenceSet.h
/// \brief contains the ReferenceSet class
///
/// $Revision: 697 $
/// \author Gerald Weber

#ifndef GBC_EXP_REFERENCESET_H
#define GBC_EXP_REFERENCESET_H "$Id: ReferenceSet.h 697 2012-11-05 15:57:05Z ge $"

#include "JobControl.h"

namespace gbc_exp {

using namespace gbc;

template<class _DequeTp>
class ReferenceSet: public std::map<typename _DequeTp::pair_type,int>
  {
  public:
    typedef typename _DequeTp::pair_type pair_type;
    typedef std::map<pair_type,int> map_type;

  bool Use_cheapest_symmetry;

  ReferenceSet(void): Use_cheapest_symmetry(true) {};

  ReferenceSet(bool sym): Use_cheapest_symmetry(sym) {};

  inline void reference_insert(pair_type bpn)
    {
    if (Use_cheapest_symmetry) bpn.reduce_to_smallest_symmetry();
    typename map_type::iterator it=this->find(bpn);
    if (it == this->end()) 
      {
      this->insert(typename map_type::value_type(bpn,1));
      }
    else it->second++;
    }

 inline void analyse(const _DequeTp& dq)
   {
   typename _DequeTp::const_iterator it;
   for(it=dq.begin(); it != dq.end(); ++it) this->reference_insert(*it);
   }

 inline int count(pair_type bpn) const
    {
    if (Use_cheapest_symmetry) bpn.reduce_to_smallest_symmetry();
    typename map_type::const_iterator it=this->find(bpn);
    if (it != this->end()) 
      {
      return it->second;
      }
    return int();
    }

  inline int operator[](const pair_type& bpn) const
    {return count(bpn);}

  inline int operator[](std::string nm) const
    {
    pair_type bp(nm);
    return count(bp);
    }

  friend std::ostream& operator<<(std::ostream& out, const ReferenceSet& ref)
    {
    typename ReferenceSet::const_iterator it;
    for (it=ref.begin(); it != ref.end(); it++)
      {
      out << it->first << "=" << it->second << " ";
      }
    out << std::endl;
    return out;
    }

  };
};//ends namespace
#endif
